@@ -24,10 +24,11 @@
 /* FIXME: Should use BonoboUIHandler.  */
 
 #include <config.h>
+#include <libgnome/gnome-i18n.h>
 #include <gnome.h>
 #include <bonobo.h>
-#include <gal/widgets/widget-color-combo.h>
 
+#include "gi-color-combo.h"
 #include "toolbar.h"
 #include "utils.h"
 #include "htmlcolor.h"
@@ -297,8 +298,6 @@ setup_color_combo (GtkHTMLControlData *cd)
         g_signal_connect (cd->html, "load_done", G_CALLBACK (load_done), cd);
 
 	cd->combo = color_combo_new (NULL, _("Automatic"), &color->color, color_group_fetch ("toolbar_text", cd));
-	GTK_WIDGET_UNSET_FLAGS (cd->combo, GTK_CAN_FOCUS);
-	gtk_container_forall (GTK_CONTAINER (cd->combo), unset_focus, NULL);
         g_signal_connect (cd->combo, "color_changed", G_CALLBACK (color_changed), cd);
 
 	gtk_widget_show_all (cd->combo);
@@ -567,7 +566,8 @@ static GtkWidget *
 create_style_toolbar (GtkHTMLControlData *cd)
 {
 	GtkWidget *hbox;
-
+	gchar *domain;
+	
 	hbox = gtk_hbox_new (FALSE, 0);
 
 	cd->toolbar_style = gtk_toolbar_new ();
@@ -582,7 +582,19 @@ create_style_toolbar (GtkHTMLControlData *cd)
 				    setup_font_size_option_menu (cd),
 				    NULL, NULL);
 
+	/* 
+	 * FIXME: steal textdomain temporarily from main-process,  and set it to 
+	 * GETTEXT_PACKAGE, after create the widgets, restore it.
+	 */
+	domain = g_strdup (textdomain (NULL));
+	textdomain (GETTEXT_PACKAGE);
+	
 	gnome_app_fill_toolbar_with_data (GTK_TOOLBAR (cd->toolbar_style), editor_toolbar_style_uiinfo, NULL, cd);
+
+	/* restore the stolen domain */
+	textdomain (domain);
+	g_free (domain);
+
 	gtk_toolbar_append_widget (GTK_TOOLBAR (cd->toolbar_style),
 				   setup_color_combo (cd),
 				   NULL, NULL);
@@ -603,6 +615,7 @@ create_style_toolbar (GtkHTMLControlData *cd)
 	cd->right_align_button = editor_toolbar_alignment_group[2].widget;
 
 	cd->unindent_button  = editor_toolbar_style_uiinfo [8].widget;
+	gtk_widget_set_sensitive (cd->unindent_button, gtk_html_get_paragraph_indentation (cd->html) != 0);
 	g_signal_connect (cd->html, "current_paragraph_indentation_changed",
 			  G_CALLBACK (indentation_changed), cd);
 
@@ -621,23 +634,45 @@ create_style_toolbar (GtkHTMLControlData *cd)
 	gtk_widget_show_all (hbox);
 
 	toolbar_update_format (cd);
+	GTK_WIDGET_UNSET_FLAGS (cd->toolbar_style, GTK_CAN_FOCUS);
+	gtk_container_forall (GTK_CONTAINER (cd->toolbar_style), unset_focus, NULL);
 
 	return hbox;
 }
 
+static gboolean
+toolbar_item_represents (GtkWidget *item, GtkWidget *widget)
+{
+	GtkWidget *parent;
+
+	if (item == widget)
+		return TRUE;
+
+	parent = gtk_widget_get_parent (widget);
+	while (parent) {
+		if (parent == item)
+			return TRUE;
+		parent = gtk_widget_get_parent (parent);
+	}
+
+	return FALSE;
+}
+
 static void
 toolbar_item_update_sensitivity (GtkWidget *widget, gpointer data)
 {
 	GtkHTMLControlData *cd = (GtkHTMLControlData *)data;
 	gboolean sensitive;
 
-	sensitive = ((cd->format_html && widget != cd->unindent_button)
-		     || widget == cd->paragraph_option
-		     || widget == cd->indent_button
-		     || (widget == cd->unindent_button && gtk_html_get_paragraph_indentation (cd->html))
-		     || widget == cd->left_align_button
-		     || widget == cd->center_button
-		     || widget == cd->right_align_button);
+	if (toolbar_item_represents (widget, cd->unindent_button))
+		return;
+
+	sensitive = (cd->format_html
+		     || toolbar_item_represents (widget, cd->paragraph_option)
+		     || toolbar_item_represents (widget, cd->indent_button)
+		     || toolbar_item_represents (widget, cd->left_align_button)
+		     || toolbar_item_represents (widget, cd->center_button)
+		     || toolbar_item_represents (widget, cd->right_align_button));
 
 	gtk_widget_set_sensitive (widget, sensitive);
 }
@@ -646,8 +681,8 @@ void
 toolbar_update_format (GtkHTMLControlData *cd)
 {
 	if (cd->toolbar_style)
-		gtk_container_forall (GTK_CONTAINER (cd->toolbar_style), 
-				      toolbar_item_update_sensitivity, cd);
+		gtk_container_foreach (GTK_CONTAINER (cd->toolbar_style), 
+		toolbar_item_update_sensitivity, cd);
 
 	if (cd->paragraph_option)
 		paragraph_style_option_menu_set_mode (cd->paragraph_option, 
