@@ -309,6 +309,9 @@ void               html_text_calc_text_size        (HTMLText              *t,
 						    gint                  *width,
 						    gint                  *asc,
 						    gint                  *dsc);
+gboolean           html_text_is_line_break         (PangoItem             *item,
+						    PangoLogAttr          *attrs,
+						    int                    offset);
 
 typedef HTMLObject * (* HTMLTextHelperFunc)       (HTMLText *, gint begin, gint end);
 HTMLObject *html_text_op_copy_helper    (HTMLText           *text,
